from .item import log, attachment, feature, story, title, step
from . import plugin
from ._internal import Launch
from ._data import parse
from . import attachment_type

parse()
__all__ = [
    'plugin',
    'Launch',
    'attachment_type',
    'parse'
]



import requests
import json
import os
import inspect
import json
from ._data import timestamp, Data, parse
import tempfile

temp_dir = tempfile.gettempdir()
FILE_PATH = os.path.join(temp_dir, 'test.json')
with open(FILE_PATH, 'w') as file:
    file.write(r'{"": ""}')


class Launch:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def get_enclosing_class_name(cls, func):
        '''
        Get the name of the enclosing class for a function.
        Returns None if the function is not a method.
        '''
        if inspect.ismethod(func) or inspect.isfunction(func):
            arg_names = inspect.getfullargspec(func).args
            if arg_names and arg_names[0] == 'self':
                return func.__qualname__.split('.')[0]
        return None

    @classmethod
    def items(cls) -> dict:
        with open(FILE_PATH, 'r') as file:
            return json.load(file)

    @classmethod
    def __update_items_file(cls, new_items: dict):
        with open(FILE_PATH, 'w') as file:
            json.dump(new_items, file)
    
    @classmethod
    def add_item(cls, item_name, item_id):
        current_items = cls.items()
        current_items[item_name] = item_id
        Launch.__update_items_file(current_items)
    
    @classmethod
    def get_latest_item(cls):
        current_items = cls.items()
        last_key, _ = list(current_items.items())[-1]
        return last_key

    @classmethod
    def delete_item(cls, item_name):
        current_items = cls.items()
        current_items.pop(item_name, None)
        Launch.__update_items_file(current_items)

    @classmethod
    def get_caller_name(cls):
        frame = inspect.currentframe()
        caller_frame = frame.f_back.f_back
        return caller_frame.f_code.co_name

    @classmethod
    def start_launch(cls):

        data = {
            'name': Data.launch_name,
            f'startTime': timestamp()}

        Data.read_data_file()
        if Data.base_item_data['launchUuid'] == '':
            respone = requests.post(url=f'{Data.endpoint}/launch', headers=Data.headers, json=data)
            launch_uuid = respone.json()['id']
            Data.update_data_file(launch_uuid)


    @classmethod
    def finish_launch(cls):
        requests.put(url=f'{Data.endpoint}/launch/{Data.base_item_data["launchUuid"]}/finish', headers=Data.headers, json={'endTime': timestamp()})


    @classmethod
    def create_report_item(
            cls,
            name: str,
            parent_item: str = '',
            type: str = '',
            description: str = '',
            has_stats: bool = True):

        parse()
        current_items = cls.items()
        parent = current_items[parent_item]
        if Data.base_item_data['launchUuid'] == '':
            Data.read_data_file()

        data = Data.base_item_data
        data['name'] = name
        data['type'] = type
        data['start_time'] = timestamp()
        data['description'] = description
        data['hasStats'] = has_stats

        response = requests.post(url=f'{Data.endpoint}/item/{parent}', headers=Data.headers, json=data)
        response_json = response.json()
        return response_json['id']

    @classmethod
    def finish_item(cls, item_name: str, passed: bool = True):
        status = 'passed' if passed else 'failed'
        current_items = cls.items()
        try:
            item = current_items[item_name]
            json_data= {
            'launchUuid': Data.base_item_data['launchUuid'],
            'endTime': timestamp(),
            'status': status}
            data = Data.headers
            data['Content-Type']='application/json'
            response = requests.put(url=f'{Data.endpoint}/item/{item}', headers=data, json=json_data)
            cls.delete_item(item_name)
        except:
            pass


    @classmethod
    def finish_passed_item(cls, item_name: str):
        current_items = cls.items()
        item = current_items[item_name]
        json_data= {
            'launchUuid': Data.base_item_data['launchUuid'],
            'endTime': timestamp(),
            'status': 'passed'}

        response = requests.put(url=f'{Data.endpoint}/item/{item}', headers=Data.headers, json=json_data)
        cls.delete_item(item_name)

    @classmethod
    def finish_failed_item(cls, item_name: str, message: str, reason: str):
        current_items = cls.items()
        item = current_items[item_name]
        json_data = {
            'launchUuid': Data.base_item_data['launchUuid'],
            'endTime': timestamp(),
            'status': 'failed',
            'issue': {
                'issueType': 'pb001',
                'comment': message},
            'description': reason}

        response = requests.put(url=f'{Data.endpoint}/item/{item}', headers=Data.headers, json=json_data)
        cls.delete_item(item_name)

    @classmethod
    def create_log(cls, item: str, message: str, level: str = "INFO"):
        current_items = cls.items()
        json_data = {
            "launchUuid": Data.base_item_data['launchUuid'],
            "itemUuid": current_items[item],
            "time": timestamp(),
            "message": message,
            "level": level,
        }

        requests.post(url=f'{Data.endpoint}/log', headers=Data.headers, json=json_data)

    @classmethod
    def add_attachment(cls, message: str, level: str, attachment: str, attachment_type: str, item: str = ''):
        file_name = os.path.basename(attachment)
        current_items = cls.items()
        item_uuid = current_items[item] if len(item) > 0 else cls.get_latest_item()
        json_body = {
            "launchUuid": Data.base_item_data['launchUuid'],
            "time": timestamp(),
            "message": message,
            "level": level,
            "itemUuid": item_uuid,
            "file": {"name": file_name}}

        data = b''
        data += b'--boundary-string\r\n'
        data += f'Content-Disposition: form-data; name="json_request_part"\r\n'.encode('utf-8')
        data += b'Content-Type: application/json\r\n\r\n'
        data += json.dumps([json_body]).encode('utf-8')
        data += b'\r\n--boundary-string\r\n'
        data += f'Content-Disposition: form-data; name="{file_name}"; filename="Screenshot.png"\r\n'.encode('utf-8')
        data += f'Content-Type: {attachment_type}\r\n\r\n'.encode('utf-8')
        with open(attachment, 'rb') as f:
            image_data = f.read()
        data += image_data
        data += b'\r\n--boundary-string--\r\n'
        headers = Data.headers
        headers['Content-Type'] = 'multipart/form-data; boundary=boundary-string'
        response = requests.post(url=f'{Data.endpoint}/log', headers=headers, data=data)
        return response


import json
from time import time


LAUNCH_DATA_FILE = './launch_data.json'

with open(LAUNCH_DATA_FILE, 'w') as file:
    file.write(r'{"launchUuid": ""}')

def timestamp():
    return str(int(time() * 1000))

class Data:
    endpoint = ''
    launch_name = ''
    uuid = ''
    project = ''
    headers = {
        'Authorization': f'Bearer {uuid}'
    }

    base_item_data = {
       'name': 'My Test Suite',
       'type': 'suite',
       'start_time': timestamp(),
       'launchUuid': ''
    }

    @classmethod
    def update_url(cls):
        cls.endpoint = f'{cls.endpoint}/api/v1/{cls.project}'
        cls.update_headers()

    @classmethod
    def update_headers(cls):
        cls.headers = {
            'Authorization': f'Bearer {cls.uuid}'}

    @classmethod
    def read_data_file(cls):
        with open(LAUNCH_DATA_FILE, 'r') as file:
            data = json.load(file)
            cls.base_item_data['launchUuid'] = data['launchUuid']

    @classmethod
    def update_data_file(cls, new_uuid):
        cls.read_data_file()
        cls.base_item_data['launchUuid'] = new_uuid
        with open(LAUNCH_DATA_FILE, 'w') as file:
            json.dump(cls.base_item_data, file)
            
def parse():
    import configparser
    import os

    filename = 'report.properties'
    config = configparser.ConfigParser()
    for root, dirs, files in os.walk('.'):
        if filename in files:
            filepath = os.path.join(root, filename)
            config.read(filepath)

    endpoint = config.get('Data', 'endpoint')
    uuid = config.get('Data', 'uuid')
    launch_name = config.get('Data', 'launch_name')
    project = config.get('Data', 'project')
    Data.endpoint = endpoint
    Data.uuid = uuid
    Data.launch_name = launch_name
    Data.project = project
    Data.update_url()

parse()



from ._internal import Launch
import functools
from typing import Any, Union, Literal
from .core import *
import traceback


class step:
    def __init__(self, name: str = None):
        self.name = name if name else None

    def __call__(self, func):
        __tracebackhide__ = True
        func.__new_name__ = self.name if self.name else func.__name__
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            __tracebackhide__ = True
            func_is_fixture = _is_fixture(func, args, kwargs)
            if not func_is_fixture:
                caller = Launch.get_caller_name()
                if 'test_' in caller or '_test' in caller:
                    parent = f'{caller}_Execution'
                else:
                    parent = caller

                item_id = Launch.create_report_item(
                    name=self.name,
                    parent_item=parent,
                    type='step',
                    has_stats=False,
                    description=func.__doc__)
    
                Launch.add_item(func.__name__, item_id)

            passed = True
            try:
                result = _run_func(func, *args, **kwargs)
            except:
                passed = False
            
            if not func_is_fixture:
                Launch.finish_item(func.__name__, passed)

            return result

        return wrapper

    def __enter__(self):
        __tracebackhide__ = True
        parent = Launch.get_latest_item()
        item_id = Launch.create_report_item(
                name=self.name,
                parent_item=parent,
                type='step',
                has_stats=False,
                description='')

        Launch.add_item(self.name, item_id)

    def __exit__(self, exc_type, exc_value, tb):
        __tracebackhide__ = True

        passed = exc_type is None
        if passed:
            Launch.finish_item(self.name, passed)

        elif not passed:
            traceback_str = ''.join(traceback.format_tb(tb))
            message = f'{exc_type.__name__}: {exc_value}'
            Launch.finish_failed_item(self.name, message=message, reason=traceback_str)


class title:
    def __init__(self, name: str):
        self.name = name
        
    def __call__(self, func):
        __tracebackhide__ = True
        func.__new_name__ = self.name
        self.__name__ = func.__name__
        self.__qualname__ = func.__qualname__

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            __tracebackhide__ = True
            func.formatted_name = self.name.format(*args, **kwargs)
            result = func(*args, **kwargs)
            return result

        return wrapper
    
    def __enter__(self):
        __tracebackhide__ = True
        parent = Launch.get_caller_name()
        item_id = Launch.create_report_item(
                name=self.name,
                parent_item=parent,
                type='step',
                has_stats=False,
                description='')

        Launch.add_item(self.name, item_id)

    def __exit__(self, exc_type, exc_value, tb):
        __tracebackhide__ = True
        passed = exc_type is None
        if passed:
            Launch.finish_item(self.name, passed)

        elif not passed:
            traceback_str = ''.join(traceback.format_tb(tb))
            message = f'{exc_type.__name__}: {exc_value}'
            Launch.finish_failed_item(self.name, message=message, reason=traceback_str)

def feature(name: str):
    def actual_decorator(cls):
        __tracebackhide__ = True

        original_setup = getattr(cls, 'setup_class', None)
        original_teardown = getattr(cls, 'teardown_class', None)
        item_id = Launch.create_report_item(
            name=name,
            type='suite',
            description=cls.__doc__)

        Launch.add_item(cls.__name__, item_id)
        Launch.add_item(name, item_id)

        @classmethod
        def new_teardown_class(cls, *args, **kwargs):
            if original_teardown is not None:
                original_teardown(*args, **kwargs)

            Launch.finish_item(cls.__name__)

        cls.teardown_class = new_teardown_class
        return cls

    return actual_decorator



def story(name: str):
    def actual_decorator(cls):
        original_setup = getattr(cls, 'setup_class', None)
        original_teardown = getattr(cls, 'teardown_class', None)
        parent = _get_class_parent(cls)
        item_id = Launch.create_report_item(
                name=name,
                parent_item=parent,
                type='story',
                description=cls.__doc__)

        Launch.add_item(cls.__name__, item_id)

        @classmethod
        def new_teardown_class(cls, *args, **kwargs):
            if original_teardown is not None:
                original_teardown(*args, **kwargs)

            Launch.finish_item(cls.__name__)

        cls.teardown_class = new_teardown_class

        return cls

    return actual_decorator


def log(*, message: str, level: str = "INFO"):
    __tracebackhide__ = True
    item = Launch.get_caller_name()
    Launch.create_log(item=item, message=message, level=level)

def attachment(*, name: str, attachment: Union[str, bytes], item: str = '', attachment_type: str, level: Literal["ERROR", "INFO", "DEBUG"] = "ERROR"):
    """Add attachment to the item (test class/case/step)
    :param item: The item name (function name)
    :param name: The attachment name
    :param attachment: attachment as bytes or the path to the attachment
    :param attachment_type: The type of the attachment (i.e use report.attachment_type.PNG)
    :param level: The log level of the the attachment (i.e if an error occured and you want to attach a screenshot use "ERROR")
    """
    Launch.add_attachment(item=item, message=name, level=level, attachment=attachment, attachment_type=attachment_type)


import inspect
import asyncio
import json
from ._internal import Launch


__all__ = ['_run_func', '_get_class_parent', '_is_fixture']

def _run_func(func, *args, **kwargs):
    __tracebackhide__ = True
    if inspect.iscoroutinefunction(func):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError as e:
            if "no running event loop" in str(e):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            else:
                raise e

            return loop.run_until_complete(func(*args, **kwargs))

    return func(*args, **kwargs)


def _get_class_parent(child_class):
    __tracebackhide__ = True
    for base_class in child_class.__bases__:
        base_class_name = base_class.__name__
        
        if base_class_name in Launch.items().keys():
            return base_class_name
        
        elif len(base_class.__bases__) > 0 :
            result = _get_class_parent(base_class)
            
            if result:
                return result

    return None

def _is_fixture(func, *args, **kwargs):
    return True if 'request' in kwargs.keys() else False




import pytest
from ._data import parse
from ._internal import Launch
import traceback


been_there_key = pytest.StashKey[bool]()
done_that_key = pytest.StashKey[str]()

parse()

latest_item = ''


def pytest_addoption(parser):
    parser.addoption("--report", action="store_true")


def pytest_sessionstart(session):
    script_path = session.config.getoption("--report")
    if script_path:
        parse()
        Launch.start_launch()


def pytest_sessionfinish(session, exitstatus):
    script_path = session.config.getoption("--report")
    if script_path:
        for item in Launch.items().keys():
            Launch.finish_item(item)
        Launch.finish_launch()

@pytest.hookimpl(tryfirst=True)
def pytest_fixture_setup(fixturedef, request):
    fixture_name = getattr(fixturedef.func, '__new_name__', request.fixturename)
    if fixture_name and '_xunit_setup_class' not in fixture_name:
        parent = Launch.get_latest_item()
        item_id = Launch.create_report_item(
                name=fixture_name,
                parent_item=parent,
                type='step',
                has_stats=True,
                description='')
    
        Launch.add_item(request.fixturename, item_id)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(item, nextitem):
    test_name = getattr(item.function, '__new_name__', item.name)
    if item.name not in Launch.items() and item.parent is not None:
        item_id = Launch.create_report_item(
            name=test_name,
            parent_item=item.parent.name,
            type='scenario',
            has_stats=True,
            description='')

        Launch.add_item(item.name, item_id)
        
        item_setup_name = f'Setup'
        item_id = Launch.create_report_item(
            name=item_setup_name,
            parent_item=item.name,
            type='before_test',
            has_stats=True,
            description='')

        Launch.add_item(f'{item.name}_{item_setup_name}', item_id)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_teardown(item, nextitem):
    item_teardown_name = 'Teardown'
    item_id = Launch.create_report_item(
        name=item_teardown_name,
        parent_item=item.name,
        type='after_test',
        has_stats=True,
        description='')

    Launch.add_item(f'{item.name}_{item_teardown_name}', item_id)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_makereport(item, call):
    excinfo = call.excinfo

    if call.when == 'setup':
        run_item_teardown(f'{item.name}_Setup', excinfo)

    if call.when == 'call':
        run_item_teardown(f'{item.name}_Execution', excinfo)

    if call.when == 'teardown':
        run_item_teardown(f'{item.name}_Teardown', excinfo)


def run_item_teardown(item_name: str, excinfo):
    if excinfo is None:
        Launch.finish_passed_item(item_name)
        if 'Setup' in item_name:
            required_item = item_name.split('_Setup')
            add_item_execution(required_item[0])

    elif excinfo is not None:
        traceback_str = ''.join(traceback.format_tb(excinfo.tb))
        Launch.finish_failed_item(item_name, message=excinfo.typename, reason=traceback_str)

def add_item_execution(item_name):
    item_execution_name = 'Execution'
    item_id = Launch.create_report_item(
        name=item_execution_name,
        parent_item=item_name,
        type='step',
        has_stats=True,
        description='')

    Launch.add_item(f'{item_name}_{item_execution_name}', item_id)


import argparse
from . import _internal

def config():
    parser = argparse.ArgumentParser(description='Pytest Report Portal Wrapper')
    parser.add_argument('--host', '-H', default='localhost', help='host:ip or domain name (Example: localhost:8080)')
    parser.add_argument('--launch-name', '-l', default='new launch' ,help='New launch name ( default is "new launch" )')
    parser.add_argument('--bearer-uuid', '-u', required=True, help='Auth bearer UUID')
    parser.add_argument('--project-name', '-p', required=True, help='Report Portal project name')
    args = parser.parse_args()
    _internal.Data.endpoint = args.host
    _internal.Data.launch_name = args.launch_name
    _internal.Data.uuid = args.bearer_uuid
    _internal.Data.project = args.project_name
    _internal.Data.update_url()
    _internal.Data.update_headers()


if __name__ == '__main__':
    data = config()

